#!/usr/bin/env python3

import argparse
import collections
import contextlib
import hashlib
import io
import ipaddress
import json
import logging
import netifaces
import os
import pathlib
import signal
import socket
import subprocess
import tempfile
from urllib import parse
from typing import Iterable, Set, TextIO, Union

import jinja2


NGINX_BASE = '/etc/nginx'

IPAddress = Union[ipaddress.IPv4Address, ipaddress.IPv6Address]

logger = logging.getLogger(__name__)


class TemporaryDirectory(tempfile.TemporaryDirectory):

	def __enter__(self):
		return pathlib.Path(self.name)

	def __str__(self):
		return self.name


class StreamConfig:
	"""
	Manage an Nginx stream configuration
	"""

	def __init__(self, name, **values):
		self.name = name
		self.path = None
		self.vals = values

	def enable(self, engine, nginx_base=NGINX_BASE):
		"""
		Add the stream config
		"""
		self.path = pathlib.Path(nginx_base).joinpath(f'streams/{self.name}.conf')
		template = engine.get_template('stream.conf')
		with self.path.open('w') as conf:
			conf.writelines(template.stream(**self.vals))

	def disable(self):
		"""
		Remove the stream config
		"""
		if self.path.exists():
			self.path.unlink()


@contextlib.contextmanager
def add_core_config(engine, opts):
	"""
	Add a minimal Nginx configuration
	"""
	template = engine.get_template('nginx.conf')
	with opts.nginx_directory as _dir:
		path = pathlib.Path(_dir)
		with path.joinpath('nginx.conf').open('w') as conf:
			conf.writelines(template.stream(**opts.__dict__))
		path.joinpath('streams').mkdir(exist_ok=True)
		yield path


def parse_cmdline(argv=None):
	parser = argparse.ArgumentParser()
	parser.add_argument('--monitor-url',
		type=parse.urlparse,
		default=parse.urlparse('http://127.0.0.1/healthz'),
	)
	parser.add_argument('--cert', '--certificate',
		type=pathlib.Path,
	)
	parser.add_argument('--key',
		type=pathlib.Path,
	)
	parser.add_argument('--nginx-directory',
		type=pathlib.Path,
		default=pathlib.Path(NGINX_BASE),
	)
	parser.add_argument('--dry-run',
		action='store_true',
	)

	mutex = parser.add_mutually_exclusive_group()
	mutex.add_argument('--verbose', '-v',
		action='count',
		default=0,
	)
	mutex.add_argument('--quiet', '-q',
		action='store_true',
		default=False,
	)

	opts = parser.parse_args(argv)

	if opts.monitor_url.scheme == 'https':
		if not opts.cert:
			parser.error(f"{opts.monitor_url!r} requires TLS certificates with --cert and --key")
		elif not opts.cert.exists():
			parser.error(f"{opts.cert} does not exist")
		elif opts.key and not opts.key.exists():
			parser.error(f"{opts.key} does not exist")
	elif opts.cert or opts.key:
		parser.error("--cert and --key are not valid for {opts.monitor_url!r}")

	if opts.dry_run:
		opts.nginx_directory = TemporaryDirectory()

	# Set verbosity
	logging.root.setLevel(
		logging.DEBUG   if opts.verbose >= 1 else
		logging.INFO    if not opts.quiet else
		logging.WARNING
	)

	return opts


def parse_annotation(annotation):
	parser = argparse.ArgumentParser()
	parser.add_argument('--proxy-all',
		action='store_true',
		default=False,
	)
	parser.add_argument('--proxy',
		action='append',
		default=[],
	)
	parser.add_argument('names',
		nargs='+',
	)

	def split(string):
		if ':' in string:
			proto, port = string.split(':')
			return proto.upper(), int(port)
		return string.lower()

	opts = parser.parse_args(annotation.split())
	opts.names = set(opts.names)
	opts.proxy = set(split(port) for port in opts.proxy)

	return opts


def main(argv=None):
	opts = parse_cmdline(argv)

	def nginx_addr_format(input):
		addr, port = input
		if addr.version == 6:
			return f'[{addr}]:{port}'
		else:
			return f'{addr}:{port}'

	# Template engine
	env = jinja2.Environment(
		loader=jinja2.FileSystemLoader('/usr/share/nginx/templates'),
		keep_trailing_newline=True, # It's a line ENDING people! Leave it in.
	)
	env.filters['nginx_addr'] = nginx_addr_format

	with contextlib.ExitStack() as stack:
		if not opts.dry_run:
			stack.callback(os.kill, os.getpid(), signal.SIGTERM)

		@stack.push
		def log_exc(exc_cls, *_):
			if exc_cls and not issubclass(exc_cls, KeyboardInterrupt):
				logger.exception("")
				return True

		# Minimal starting config
		nginx_dir = stack.enter_context(add_core_config(env, opts))
		try:
			subprocess.check_call(['nginx', '-vt'])
		except FileNotFoundError:
			logger.warning(
					"no nginx binary found, not checking configuration files")
		logger.info("generated minimal nginx configuration in %s", nginx_dir)

		# Begin monitoring k8s API for NodePort services
		cmd = ['kubectl', 'get',  'services', '--all-namespaces',
				'--watch', '--output', 'json']
		proc = stack.enter_context(
			subprocess.Popen(cmd, stdout=subprocess.PIPE, text=True))
		stack.callback(proc.terminate)
		logger.info("pid %s: kubectl", proc.pid)

		if not opts.dry_run and os.fork():
			# Parent becomes nginx
			logger.info("pid %s: nginx", os.getpid())
			os.execvp('nginx', ['nginx'])

		logger.info("pid %s: monitor", os.getpid())

		streams_map = collections.defaultdict(list)
		for uid, stream in filter_services(proc.stdout):
			if not uid:
				logging.info("reloading nginx")
				if not opts.dry_run:
					subprocess.check_call(['nginx', '-s', 'reload'])
				continue

			streams = streams_map[uid]
			if stream:
				stream.enable(env, nginx_base=nginx_dir)
				streams.append(stream)
				logging.info("stream added: %s", stream.name)
			else:
				for stream in streams:
					stream.disable()
					logging.info("stream removed: %s", stream.name)
				del streams[:]
				logging.info("service cleared: %s", uid)


def filter_services(stream: TextIO):
	"""
	Filter Service objects from stream into (unique-ID, StreamConfig) values

	Parses Kubernetes Service objects from a JSON stream and filters them, 
	yielding (unique-ID, StreamConfig) tuples. See `node_ports` for the 
	filtering information.

	The yielded unique-ID (uid) values are the Service uids from the Kubernetes 
	API.

	If StreamConfig value is None (and the uid is not None) all the stream 
	configurations for the service with that UID are cleared.

	If the uid is None, it indicates that a service has been completely handled, 
	and any configuration changes should be committed.
	"""
	external_addrs = set(get_host_addr())
	buf = io.StringIO()
	for line in stream:
		buf.write(line)
		if line and line != '}\n':
			continue
		buf.seek(0)
		service = json.load(buf)
		buf.seek(0)
		buf.truncate()
		yield from node_ports(service, external_addrs)
		yield None, None  # Nginx reload
		if stream.closed:
			logger.info("kubectl stream has closed")
			return
		if not line:
			logger.warning("empty line from an open stream")


def node_ports(service: dict, host_addrs: Set[IPAddress]):
	"""
	Yield port information from services relevant to the current node's host

	Returns a generator that yields (unique-ID, StreamConfig) values for 
	any of the node host's addresses listed in the 'kodo.org.uk/port-proxy' 
	annotation, if a service is eligible (type == NodePort AND 
	'kodo.org.uk/port-proxy' in .metadata.annotations).

	See `filter_services` for information on the yielded values.
	"""

	metadata = service['metadata']
	spec = service['spec']
	name = f"{metadata['name']}.{metadata['namespace']}"
	uid = metadata['uid']
	logger.info("got a service: %s (%s)", name, uid)

	# Clear old streams for this service
	yield uid, None

	# TODO: On addressing #20, remove NodePort restriction
	if spec.get('type') != 'NodePort':
		logger.debug("rejecting service: not a NodePort")
		return
	try:
		ports = spec['ports']
		# TODO: Awaiting #20
		# cluster_ip = ipaddress.ip_address(spec['clusterIP'])
		proxy_args = metadata['annotations']['kodo.org.uk/port-proxy']
	except KeyError as exc:
		logger.debug("rejecting service: missing %s", exc)
		return

	logger.debug("kodo.org.uk/port-proxy: %s", proxy_args)
	opts = parse_annotation(proxy_args)

	addresses = set(addr_from_names(opts.names))
	addresses.intersection_update(host_addrs)
	if not addresses:
		logger.debug("rejecting service: no available addresses")
		return

	for addr in addresses:
		for port in ports:
			logger.debug("generating for %s and %s", addr, port['name'])
			yield uid, StreamConfig(
				name=f"{name}.{port['name']}",
				protocol=port['protocol'],
				external_port=(addr, port['port']),
				# TODO: Awaiting #20
				# internal_port=(cluster_ip, port['port']),
				internal_port=(addr, port['nodePort']),
				use_proxy_protocol = (
					opts.proxy_all or
					port['name'] in opts.proxy or
					(port['protocol'], port['port']) in opts.proxy
				),
			)


def get_host_addr() -> Iterable[IPAddress]:
	"""
	Yield all IP addresses configured on the host
	"""
	for iface in netifaces.interfaces():
		for af, iface_addr in netifaces.ifaddresses(iface).items():
			if af not in {socket.AF_INET, socket.AF_INET6}:
				continue
			for addr in iface_addr:
				try:
					yield ipaddress.ip_address(addr['addr'])
				except ValueError:
					if '%' not in addr['addr']:
						raise
					logger.warning("ignoring scoped IPv6 address")


def addr_from_names(names: Iterable[str]) -> Iterable[IPAddress]:
	"""
	Generate IPAddress objects from a list of addresses & names

	Returns a generator
	"""
	for name in names:
		try:
			yield ipaddress.ip_address(name)
		except ValueError:
			for af, *_, addrinfo in socket.getaddrinfo(name, None):
				if af not in {socket.AF_INET, socket.AF_INET6}:
					continue
				yield ipaddress.ip_address(addrinfo[0])


if __name__ == '__main__':
	logging.basicConfig()
	logger = logging.getLogger('monitor')
	try:
		main()
	except KeyboardInterrupt:
		logger.info("exit on SIGINT")
	except:
		logger.exception("error from monitor process")
